@@ -1,9 +1,10 @@
 """System Bridge Connector: WebSocket Client"""
 from __future__ import annotations
 
+import asyncio
 from collections.abc import Callable
-import json
 import socket
+from uuid import uuid4
 
 import aiohttp
 
@@ -17,7 +18,7 @@ from systembridgeconnector.const import (
     EVENT_DIRECTORIES,
     EVENT_EVENT,
     EVENT_FILE,
-    EVENT_FILES,
+    EVENT_ID,
     EVENT_KEY,
     EVENT_MESSAGE,
     EVENT_MODULE,
@@ -31,19 +32,13 @@ from systembridgeconnector.const import (
     EVENT_URL,
     EVENT_VERSION,
     MODEL_MAP,
-    MODEL_MEDIA_DIRECTORIES,
-    MODEL_MEDIA_FILE,
-    MODEL_MEDIA_FILES,
     MODEL_RESPONSE,
     SUBTYPE_BAD_API_KEY,
     SUBTYPE_LISTENER_ALREADY_REGISTERED,
     TYPE_APPLICATION_UPDATE,
     TYPE_DATA_UPDATE,
-    TYPE_DIRECTORIES,
     TYPE_ERROR,
     TYPE_EXIT_APPLICATION,
-    TYPE_FILE,
-    TYPE_FILES,
     TYPE_GET_DATA,
     TYPE_GET_DIRECTORIES,
     TYPE_GET_FILE,
@@ -66,6 +61,10 @@ from systembridgeconnector.exceptions import (
     ConnectionClosedException,
     ConnectionErrorException,
 )
+from systembridgeconnector.models.media_directories import MediaDirectories
+from systembridgeconnector.models.media_files import File as MediaFile, MediaFiles
+from systembridgeconnector.models.request import Request
+from systembridgeconnector.models.response import Response
 
 
 class WebSocketClient(Base):
@@ -82,6 +81,7 @@ class WebSocketClient(Base):
         self._api_host = api_host
         self._api_port = api_port
         self._api_key = api_key
+        self._response_futures: dict[str, asyncio.Future] = {}
         self._session: aiohttp.ClientSession | None = None
         self._websocket: aiohttp.ClientWebSocketResponse | None = None
 
@@ -90,6 +90,25 @@ class WebSocketClient(Base):
         """Get connection state."""
         return self._websocket is not None and not self._websocket.closed
 
+    async def _send_message(
+        self,
+        request: Request,
+    ) -> Response:
+        """Send a message to the WebSocket"""
+        if not self.connected or self._websocket is None:
+            raise ConnectionClosedException("Connection is closed")
+
+        request.api_key = self._api_key
+        request.id = uuid4().hex
+        future: asyncio.Future[Response] = asyncio.get_running_loop().create_future()
+        self._response_futures[request.id] = future
+        await self._websocket.send_str(request.json())
+        self._logger.debug("Sent message: %s", request.json(exclude={EVENT_API_KEY}))
+        try:
+            return await future
+        finally:
+            self._response_futures.pop(request.id)
+
     async def close(self) -> None:
         """Close connection"""
         self._logger.info("Closing WebSocket connection")
@@ -127,182 +146,136 @@ class WebSocketClient(Base):
     async def application_update(
         self,
         version: str | None = None,
-    ) -> None:
+    ) -> Response:
         """Update application"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Updating application")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_APPLICATION_UPDATE,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_VERSION: version,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_APPLICATION_UPDATE,
+                    EVENT_VERSION: version,
+                }
             )
+        )
 
-    async def exit_backend(self) -> None:
+    async def exit_backend(self) -> Response:
         """Exit backend"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Exiting backend")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_EXIT_APPLICATION,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_EXIT_APPLICATION,
+                }
             )
+        )
 
     async def get_data(
         self,
         modules: list[str],
-    ) -> None:
+    ) -> Response:
         """Get data from server"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Getting data from server: %s", modules)
-
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_GET_DATA,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_MODULES: modules,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_GET_DATA,
+                    EVENT_MODULES: modules,
+                }
             )
+        )
 
-    async def get_directories(self) -> None:
+    async def get_directories(self) -> MediaDirectories:
         """Get directories"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Getting directories:")
-
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_GET_DIRECTORIES,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        response = await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_GET_DIRECTORIES,
+                }
             )
+        )
+        return MediaDirectories(**response.dict()[EVENT_DIRECTORIES])
 
     async def get_files(
         self,
         base: str,
         path: str | None = None,
-    ) -> None:
+    ) -> MediaFiles:
         """Get files"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Getting files: %s - %s", base, path)
-
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_GET_FILES,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_BASE: base,
-                        EVENT_PATH: path,
-                    }
-                )
+        response = await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_GET_FILES,
+                    EVENT_BASE: base,
+                    EVENT_PATH: path,
+                }
             )
+        )
+        return MediaFiles(**response.dict()[EVENT_DIRECTORIES])
 
     async def get_file(
         self,
         base: str,
         path: str,
-    ) -> None:
+    ) -> MediaFile:
         """Get files"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Getting file: %s - %s", base, path)
-
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_GET_FILE,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_BASE: base,
-                        EVENT_PATH: path,
-                    }
-                )
+        response = await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_GET_FILE,
+                    EVENT_BASE: base,
+                    EVENT_PATH: path,
+                }
             )
+        )
+        return MediaFile(**response.dict()[EVENT_FILE])
 
     async def register_data_listener(
         self,
         modules: list[str],
-    ) -> None:
+    ) -> Response:
         """Register data listener"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Registering data listener: %s", modules)
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_REGISTER_DATA_LISTENER,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_MODULES: modules,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_REGISTER_DATA_LISTENER,
+                    EVENT_MODULES: modules,
+                }
             )
+        )
 
     async def keyboard_keypress(
         self,
         key: str,
-    ) -> None:
+    ) -> Response:
         """Keyboard keypress"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Press key: %s", key)
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_KEYBOARD_KEYPRESS,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_KEY: key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_KEYBOARD_KEYPRESS,
+                    EVENT_KEY: key,
+                }
             )
+        )
 
     async def keyboard_text(
         self,
         text: str,
-    ) -> None:
+    ) -> Response:
         """Keyboard keypress"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Enter text: %s", text)
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_KEYBOARD_TEXT,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_TEXT: text,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_KEYBOARD_TEXT,
+                    EVENT_TEXT: text,
+                }
             )
+        )
 
     async def send_notification(
         self,
@@ -311,162 +284,117 @@ class WebSocketClient(Base):
         app_name: str | None = None,
         app_icon: str | None = None,
         timeout: int = 5,
-    ) -> None:
+    ) -> Response:
         """Send notification"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Send notification: %s", message)
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_NOTIFICATION,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_MESSAGE: message,
-                        EVENT_TITLE: title,
-                        EVENT_APP_NAME: app_name,
-                        EVENT_APP_ICON: app_icon,
-                        EVENT_TIMEOUT: timeout,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_NOTIFICATION,
+                    EVENT_MESSAGE: message,
+                    EVENT_TITLE: title,
+                    EVENT_APP_NAME: app_name,
+                    EVENT_APP_ICON: app_icon,
+                    EVENT_TIMEOUT: timeout,
+                }
             )
+        )
 
     async def open_path(
         self,
         path: str,
-    ) -> None:
+    ) -> Response:
         """Open path"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Opening path: %s", path)
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_OPEN,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_PATH: path,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_OPEN,
+                    EVENT_PATH: path,
+                }
             )
+        )
 
     async def open_url(
         self,
         url: str,
-    ) -> None:
+    ) -> Response:
         """Open url"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Opening URL: %s", url)
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_OPEN,
-                        EVENT_API_KEY: self._api_key,
-                        EVENT_URL: url,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_OPEN,
+                    EVENT_URL: url,
+                }
             )
+        )
 
-    async def power_sleep(self) -> None:
+    async def power_sleep(self) -> Response:
         """Power sleep"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Power sleep")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_POWER_SLEEP,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_POWER_SLEEP,
+                }
             )
+        )
 
-    async def power_hibernate(self) -> None:
+    async def power_hibernate(self) -> Response:
         """Power hibernate"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Power hibernate")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_POWER_HIBERNATE,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_POWER_HIBERNATE,
+                }
             )
+        )
 
-    async def power_restart(self) -> None:
+    async def power_restart(self) -> Response:
         """Power restart"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Power restart")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_POWER_RESTART,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_POWER_RESTART,
+                }
             )
+        )
 
-    async def power_shutdown(self) -> None:
+    async def power_shutdown(self) -> Response:
         """Power shutdown"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Power shutdown")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_POWER_SHUTDOWN,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_POWER_SHUTDOWN,
+                }
             )
+        )
 
-    async def power_lock(self) -> None:
+    async def power_lock(self) -> Response:
         """Power lock"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Power lock")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_POWER_LOCK,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_POWER_LOCK,
+                }
             )
+        )
 
-    async def power_logout(self) -> None:
+    async def power_logout(self) -> Response:
         """Power logout"""
-        if not self.connected:
-            raise ConnectionClosedException("Connection is closed")
-
         self._logger.info("Power logout")
-        if self._websocket is not None:
-            await self._websocket.send_str(
-                json.dumps(
-                    {
-                        EVENT_EVENT: TYPE_POWER_LOGOUT,
-                        EVENT_API_KEY: self._api_key,
-                    }
-                )
+        return await self._send_message(
+            Request(
+                **{
+                    EVENT_EVENT: TYPE_POWER_LOGOUT,
+                }
             )
+        )
 
     async def listen(
         self,
@@ -478,6 +406,25 @@ class WebSocketClient(Base):
         async def _callback_message(message: dict) -> None:
             """Message Callback"""
             self._logger.debug("New message: %s", message[EVENT_TYPE])
+
+            if message.get(EVENT_ID) is not None:
+                future = self._response_futures.get(message[EVENT_ID])
+                if future is not None:
+                    response = Response(**message)
+                    self._logger.info(
+                        "Response: %s",
+                        response.json(
+                            include={
+                                EVENT_ID,
+                                EVENT_TYPE,
+                                EVENT_SUBTYPE,
+                                EVENT_MESSAGE,
+                            },
+                            exclude_unset=True,
+                        ),
+                    )
+                    future.set_result(response)
+
             if message[EVENT_TYPE] == TYPE_ERROR:
                 if message[EVENT_SUBTYPE] == SUBTYPE_LISTENER_ALREADY_REGISTERED:
                     self._logger.debug(message)
@@ -498,41 +445,6 @@ class WebSocketClient(Base):
                         message[EVENT_MODULE],
                         model(**message[EVENT_DATA]),
                     )
-            elif message[EVENT_TYPE] == TYPE_DIRECTORIES:
-                self._logger.debug("New directories")
-                model = MODEL_MAP[MODEL_MEDIA_DIRECTORIES]
-                await callback(
-                    MODEL_MEDIA_DIRECTORIES,
-                    model(
-                        **{
-                            EVENT_DIRECTORIES: message[EVENT_DIRECTORIES],
-                        }
-                    ),
-                )
-            elif message[EVENT_TYPE] == TYPE_FILES:
-                self._logger.debug("New files")
-                model = MODEL_MAP[MODEL_MEDIA_FILES]
-                await callback(
-                    MODEL_MEDIA_FILES,
-                    model(
-                        **{
-                            EVENT_FILES: message[EVENT_FILES],
-                            EVENT_PATH: message[EVENT_PATH],
-                        }
-                    ),
-                )
-            elif message[EVENT_TYPE] == TYPE_FILE:
-                self._logger.debug("New file")
-                model = MODEL_MAP[MODEL_MEDIA_FILE]
-                await callback(
-                    MODEL_MEDIA_FILE,
-                    model(
-                        **{
-                            EVENT_FILE: message[EVENT_FILE],
-                            EVENT_PATH: message[EVENT_PATH],
-                        }
-                    ),
-                )
             else:
                 self._logger.debug("Other message: %s", message[EVENT_TYPE])
                 if accept_other_types:
